/*====================================================================*
 * fits.c
 *
 * Copyright(c) 2022 TsukiTsuki. (http://tsukitsuki.tips/)
 *====================================================================*/
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/statvfs.h>
#include <string.h>
#include <math.h>
#include <regex.h>
#include "fits.h"

double GetTime();

/*--------------------------------------------------------------------*
 *  ϑ𐶐
 *--------------------------------------------------------------------*/
char *
CreateObsTime()
{
    static  char    obs_time[64];
    time_t  stime;
    struct  tm *now;
    struct  timeval tv;
    int     yy, mm, dd, h, m, s;
     
    stime = time(NULL);
    now = localtime(&stime);
    gettimeofday(&tv, NULL);
 
    yy = (now->tm_year + 1900) % 2000;
    mm = now->tm_mon + 1;
    dd = now->tm_mday;
    h  = now->tm_hour;
    m  = now->tm_min;
    s  = now->tm_sec;

    sprintf(obs_time, "%04d-%02d-%02dT%02d:%02d:%02d.%06d+09:00", yy, mm, dd, h, m, s, tv.tv_usec);

    return obs_time;
}

/*--------------------------------------------------------------------*
 *  16bit̃GfBAύX
 *--------------------------------------------------------------------*/
void EndianChange(unsigned char *dst, unsigned char *src, int srcSize, int bpp){
	switch(bpp){
		case 16:
		case 10:
		case 12:
			for(int i = 0; i < srcSize; i+=2){
				dst[i]   = src[i+1];
				dst[i+1] = src[i];
			}
			break;
		case 8:
		default:
			break;
	}
}

/*--------------------------------------------------------------------*
 *  DatpXZTo
 *--------------------------------------------------------------------*/
void GetSensorName(char *src, char *dst){
    *dst = '\0';
    const char *regex = "(imx[0-9]+)";
    regex_t regexBuffer;

    if(regcomp(&regexBuffer, regex, REG_EXTENDED | REG_NEWLINE) != 0){
        return;
    }

    #define Size 4
    regmatch_t patternMatch[Size];
    if(regexec(&regexBuffer, src, Size, patternMatch, 0) != 0){
	return;
    }

    int start = patternMatch[1].rm_so;
    int length= patternMatch[1].rm_eo - patternMatch[1].rm_so;
    strncpy(dst, src + start, length);
    *(dst+length) = '\0';
    regfree(&regexBuffer);
}



int FtsSize;
unsigned char *FtsImage;
char *FtsHeader;
char sensorname[18];

/*--------------------------------------------------------------------*
 *  FITSt@Cۑp̊ef[^
 *--------------------------------------------------------------------*/
int InitFits(int BinSize, int width, int height, char *fname)
{
    if(BinSize % FITS_BLOCK_SIZE > 0)
        FtsSize = BinSize / FITS_BLOCK_SIZE + 1;
    else
        FtsSize = BinSize / FITS_BLOCK_SIZE;
    FtsSize = FtsSize * FITS_BLOCK_SIZE;

    FtsImage = (unsigned char *)malloc(FtsSize);
    if(FtsImage == NULL){
        perror("malloc error");
        exit(EXIT_FAILURE);
    }

    GetSensorName(fname, sensorname);

    FtsHeader = (char *)malloc(FITS_BLOCK_SIZE);

    int offset = 0;
    sprintf(FtsHeader+offset,"%-80s", "SIMPLE  =                   T / Created by SensorLaunch");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"BITPIX  = %19d / %-49s", 16, "number of bits per data pixel");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"NAXIS   = %19d / %-49s", 2, "number of data axes");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"NAXIS1  = %19d / %-49s", width, "length of data axis 1");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"NAXIS2  = %19d / %-49s", height, "length of data axis 2");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"BZERO   = %19d / %-49s", 0, "data range offset");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"BSCALE  = %19d / %-49s", 1, "lscaling factor");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"SENSOR  = '%-17s' / %-49s", sensorname, "image sensor name");
    offset += FITS_RECORD_SIZE;
    sprintf(FtsHeader+offset,"%-80s", "END");
    for(int i=0; i<FITS_BLOCK_SIZE; i++){
      if(*(FtsHeader+i) == 0x00)
        *(FtsHeader+i) = 0x20;
    }

    return 0;
}

/*--------------------------------------------------------------------*
 *  FITSt@Cɏo
 *--------------------------------------------------------------------*/
int WriteFTS(unsigned char *BinImage, int BinSize, int bpp, char *wfname)
{
    EndianChange(FtsImage, BinImage, BinSize, bpp);

    int w;
    FILE *fpFts;

    fpFts = fopen(wfname, "wb");
    w = fwrite(FtsHeader, sizeof(char), FITS_BLOCK_SIZE, fpFts);
    printf("%ld ", FITS_BLOCK_SIZE * sizeof(char));
    w = fwrite(FtsImage, sizeof(char), FtsSize, fpFts);
    printf("+ %d bytes\n", w);
    fclose(fpFts);

    return 0;
}
