// g++ -O3 -fopenmp -o myappV13 myappV13.c `pkg-config --cflags --libs opencv`
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <linux/videodev2.h>
#include <linux/v4l2-controls.h>
#include <iostream>
#include "opencv2/opencv.hpp"
#include "opencv2/highgui/highgui.hpp"

#define CLEAR(x) memset(&(x), 0, sizeof(x))
#define RETRY   5

#define TEGRA_CAMERA_CID_BASE   (V4L2_CTRL_CLASS_CAMERA | 0x2000)

#define TEGRA_CAMERA_CID_FRAME_LENGTH           (TEGRA_CAMERA_CID_BASE+0)
#define TEGRA_CAMERA_CID_COARSE_TIME            (TEGRA_CAMERA_CID_BASE+1)
#define TEGRA_CAMERA_CID_COARSE_TIME_SHORT      (TEGRA_CAMERA_CID_BASE+2)
#define TEGRA_CAMERA_CID_GROUP_HOLD             (TEGRA_CAMERA_CID_BASE+3)
#define TEGRA_CAMERA_CID_HDR_EN                 (TEGRA_CAMERA_CID_BASE+4)
#define TEGRA_CAMERA_CID_EEPROM_DATA            (TEGRA_CAMERA_CID_BASE+5)
#define TEGRA_CAMERA_CID_OTP_DATA               (TEGRA_CAMERA_CID_BASE+6)
#define TEGRA_CAMERA_CID_FUSE_ID                (TEGRA_CAMERA_CID_BASE+7)
#define TEGRA_CAMERA_CID_SENSOR_MODE_ID         (TEGRA_CAMERA_CID_BASE+8)

#define TEGRA_CAMERA_CID_GAIN                   (TEGRA_CAMERA_CID_BASE+9)
#define TEGRA_CAMERA_CID_EXPOSURE               (TEGRA_CAMERA_CID_BASE+10)
#define TEGRA_CAMERA_CID_FRAME_RATE             (TEGRA_CAMERA_CID_BASE+11)
#define TEGRA_CAMERA_CID_EXPOSURE_SHORT         (TEGRA_CAMERA_CID_BASE+12)

#define TEGRA_CAMERA_CID_SENSOR_CONFIG          (TEGRA_CAMERA_CID_BASE+50)
#define TEGRA_CAMERA_CID_SENSOR_MODE_BLOB       (TEGRA_CAMERA_CID_BASE+51)
#define TEGRA_CAMERA_CID_SENSOR_CONTROL_BLOB    (TEGRA_CAMERA_CID_BASE+52)

#define TEGRA_CAMERA_CID_VI_BYPASS_MODE         (TEGRA_CAMERA_CID_BASE+100)
#define TEGRA_CAMERA_CID_OVERRIDE_ENABLE        (TEGRA_CAMERA_CID_BASE+101)
#define TEGRA_CAMERA_CID_VI_HEIGHT_ALIGN        (TEGRA_CAMERA_CID_BASE+102)
#define TEGRA_CAMERA_CID_VI_SIZE_ALIGN          (TEGRA_CAMERA_CID_BASE+103)
#define TEGRA_CAMERA_CID_WRITE_ISPFORMAT        (TEGRA_CAMERA_CID_BASE+104)

#define TEGRA_CAMERA_CID_SENSOR_SIGNAL_PROPERTIES  (TEGRA_CAMERA_CID_BASE+105)
#define TEGRA_CAMERA_CID_SENSOR_IMAGE_PROPERTIES   (TEGRA_CAMERA_CID_BASE+106)
#define TEGRA_CAMERA_CID_SENSOR_CONTROL_PROPERTIES (TEGRA_CAMERA_CID_BASE+107)
#define TEGRA_CAMERA_CID_SENSOR_DV_TIMINGS         (TEGRA_CAMERA_CID_BASE+108)
#define TEGRA_CAMERA_CID_LOW_LATENCY         (TEGRA_CAMERA_CID_BASE+109)

int ctl_ids[]={
    TEGRA_CAMERA_CID_FRAME_LENGTH               ,  // 0x009a2000
    TEGRA_CAMERA_CID_COARSE_TIME                ,  // 0x009a2001
    TEGRA_CAMERA_CID_COARSE_TIME_SHORT          ,  // 0x009a2002
    TEGRA_CAMERA_CID_GROUP_HOLD                 ,  // 0x009a2003
    TEGRA_CAMERA_CID_HDR_EN                     ,  // 0x009a2004
    TEGRA_CAMERA_CID_EEPROM_DATA                ,  // 0x009a2005
    TEGRA_CAMERA_CID_OTP_DATA                   ,  // 0x009a2006
    TEGRA_CAMERA_CID_FUSE_ID                    ,  // 0x009a2007
    TEGRA_CAMERA_CID_SENSOR_MODE_ID             ,  // 0x009a2008
    TEGRA_CAMERA_CID_GAIN                       ,  // 0x009a2009
    TEGRA_CAMERA_CID_EXPOSURE                   ,  // 0x009a200A
    TEGRA_CAMERA_CID_FRAME_RATE                 ,  // 0x009a200B
    TEGRA_CAMERA_CID_EXPOSURE_SHORT             ,  // 0x009a200C
    TEGRA_CAMERA_CID_SENSOR_CONFIG              ,  // 0x009a2032
    TEGRA_CAMERA_CID_SENSOR_MODE_BLOB           ,  // 0x009a2033
    TEGRA_CAMERA_CID_SENSOR_CONTROL_BLOB        ,  // 0x009a2034
    TEGRA_CAMERA_CID_VI_BYPASS_MODE             ,  // 0x009a2064
    TEGRA_CAMERA_CID_OVERRIDE_ENABLE            ,  // 0x009a2065
    TEGRA_CAMERA_CID_VI_HEIGHT_ALIGN            ,  // 0x009a2066
    TEGRA_CAMERA_CID_VI_SIZE_ALIGN              ,  // 0x009a2067
    TEGRA_CAMERA_CID_WRITE_ISPFORMAT            ,  // 0x009a2068
    TEGRA_CAMERA_CID_SENSOR_SIGNAL_PROPERTIES   ,  // 0x009a2069
    TEGRA_CAMERA_CID_SENSOR_IMAGE_PROPERTIES    ,  // 0x009a206A
    TEGRA_CAMERA_CID_SENSOR_CONTROL_PROPERTIES  ,  // 0x009a206B
    TEGRA_CAMERA_CID_SENSOR_DV_TIMINGS          ,  // 0x009a206C
    TEGRA_CAMERA_CID_LOW_LATENCY                ,  // 0x009a206D
    V4L2_CID_BRIGHTNESS                         ,  // 0x00980900
    V4L2_CID_CONTRAST                           ,  // 0x00980901
    V4L2_CID_SATURATION                         ,  // 0x00980902
    V4L2_CID_HUE                                ,  // 0x00980903
    V4L2_CID_AUDIO_VOLUME                       ,  // 0x00980905
    V4L2_CID_AUDIO_BALANCE                      ,  // 0x00980906
    V4L2_CID_AUDIO_BASS                         ,  // 0x00980907
    V4L2_CID_AUDIO_TREBLE                       ,  // 0x00980908
    V4L2_CID_AUDIO_MUTE                         ,  // 0x00980909
    V4L2_CID_AUDIO_LOUDNESS                     ,  // 0x0098090A
    V4L2_CID_BLACK_LEVEL                        ,  // 0x0098090B
    V4L2_CID_AUTO_WHITE_BALANCE                 ,  // 0x0098090C
    V4L2_CID_DO_WHITE_BALANCE                   ,  // 0x0098090D
    V4L2_CID_RED_BALANCE                        ,  // 0x0098090E
    V4L2_CID_BLUE_BALANCE                       ,  // 0x0098090F
    V4L2_CID_GAMMA                              ,  // 0x00980910
    V4L2_CID_EXPOSURE                           ,  // 0x00980911
    V4L2_CID_AUTOGAIN                           ,  // 0x00980912
    V4L2_CID_GAIN                               ,  // 0x00980913
    V4L2_CID_HFLIP                              ,  // 0x00980914
    V4L2_CID_VFLIP                              ,  // 0x00980915
    V4L2_CID_POWER_LINE_FREQUENCY               ,  // 0x00980918
    V4L2_CID_EXPOSURE_AUTO                      ,  // 0x009a0901
    V4L2_CID_EXPOSURE_ABSOLUTE                  ,  // 0x009a0902
    V4L2_CID_EXPOSURE_AUTO_PRIORITY             ,  // 0x009a0903
    V4L2_CID_PAN_RELATIVE                       ,  // 0x009a0904
    V4L2_CID_TILT_RELATIVE                      ,  // 0x009a0905
    V4L2_CID_PAN_RESET                          ,  // 0x009a0906
    V4L2_CID_TILT_RESET                         ,  // 0x009a0907
    V4L2_CID_PAN_ABSOLUTE                       ,  // 0x009a0908
    V4L2_CID_TILT_ABSOLUTE                      ,  // 0x009a0909
    V4L2_CID_FOCUS_ABSOLUTE                     ,  // 0x009a090A
    V4L2_CID_FOCUS_RELATIVE                     ,  // 0x009a090B
    V4L2_CID_FOCUS_AUTO                         ,  // 0x009a090C
    V4L2_CID_ZOOM_ABSOLUTE                      ,  // 0x009a090D
    V4L2_CID_ZOOM_RELATIVE                      ,  // 0x009a090E
    V4L2_CID_ZOOM_CONTINUOUS                    ,  // 0x009a090F
    V4L2_CID_PRIVACY                            ,  // 0x009a0910
    V4L2_CID_IRIS_ABSOLUTE                      ,  // 0x009a0911
    V4L2_CID_IRIS_RELATIVE                      ,  // 0x009a0912
    V4L2_CID_AUTO_EXPOSURE_BIAS                 ,  // 0x009a0913
    V4L2_CID_AUTO_N_PRESET_WHITE_BALANCE        ,  // 0x009a0914
    V4L2_CID_WIDE_DYNAMIC_RANGE                 ,  // 0x009a0915
    V4L2_CID_IMAGE_STABILIZATION                ,  // 0x009a0916
    V4L2_CID_ISO_SENSITIVITY                    ,  // 0x009a0917
    V4L2_CID_ISO_SENSITIVITY_AUTO               ,  // 0x009a0918
    V4L2_CID_EXPOSURE_METERING                  ,  // 0x009a0919
    V4L2_CID_SCENE_MODE                         ,  // 0x009a091A
    V4L2_CID_3A_LOCK                            ,  // 0x009a091B
    V4L2_CID_AUTO_FOCUS_START                   ,  // 0x009a091C
    V4L2_CID_AUTO_FOCUS_STOP                    ,  // 0x009a091D
    V4L2_CID_AUTO_FOCUS_STATUS                  ,  // 0x009a091E
    V4L2_CID_AUTO_FOCUS_RANGE                   ,  // 0x009a091F
    V4L2_CID_PAN_SPEED                          ,  // 0x009a0920
    V4L2_CID_TILT_SPEED                         ,  // 0x009a0921
    V4L2_CID_VBLANK                             ,  // 0x009e0901
    V4L2_CID_HBLANK                             ,  // 0x009e0902
    V4L2_CID_ANALOGUE_GAIN                      ,  // 0x009e0903
    V4L2_CID_TEST_PATTERN_RED                   ,  // 0x009e0904
    V4L2_CID_TEST_PATTERN_GREENR                ,  // 0x009e0905
    V4L2_CID_TEST_PATTERN_BLUE                  ,  // 0x009e0906
    V4L2_CID_TEST_PATTERN_GREENB                ,  // 0x009e0907
    V4L2_CID_LINK_FREQ                          ,  // 0x009f0901
    V4L2_CID_PIXEL_RATE                         ,  // 0x009f0902
    V4L2_CID_TEST_PATTERN                       ,  // 0x009f0903
    V4L2_CID_DEINTERLACING_MODE                 ,  // 0x009f0904
    V4L2_CID_DIGITAL_GAIN                       ,  // 0x009f0905
};

static const char * const ctl_type1[]{
        "INTEGER",     
        "BOOLEAN",     
        "MENU",        
        "BUTTON",      
        "INTEGER64",   
        "CTRL_CLASS",  
        "STRING",      
        "BITMASK",     
        "INTEGER_MENU"
};

static const char * const ctl_type2[]{
        "U8",
        "U16",
        "U32"
};



int *effective_ctl;
int controls;
int ctl_mode;
int ctl_mode_index;

unsigned long pixformat=0;
int WIDTH, HEIGHT;
//char *dev_name = "/dev/video3";
char device_name[] = "/dev/video0";
char *dev_name = device_name;
struct buffer {
        void   *start;
        size_t  length;
};
struct buffer          *buffers;
static unsigned int     n_buffers;
static int              frame_number = 0;
static unsigned int     read_buffer_size = 0;                    

IplImage *RGBImg, *BayerImg;

int pixelformat;
int pixelbit = 8;
char str[5];

int get_code()
{
    if(
       pixelformat == V4L2_PIX_FMT_SBGGR12
    || pixelformat == V4L2_PIX_FMT_SBGGR12P
    || pixelformat == V4L2_PIX_FMT_SBGGR10
    || pixelformat == V4L2_PIX_FMT_SBGGR10P
    || pixelformat == V4L2_PIX_FMT_SBGGR8
    )
        return CV_BayerBG2RGB;
    else if(
       pixelformat == V4L2_PIX_FMT_SGBRG12
    || pixelformat == V4L2_PIX_FMT_SGBRG12P
    || pixelformat == V4L2_PIX_FMT_SGBRG10
    || pixelformat == V4L2_PIX_FMT_SGBRG10P
    || pixelformat == V4L2_PIX_FMT_SGBRG8
    )
        return CV_BayerGB2RGB;
    else if(
       pixelformat == V4L2_PIX_FMT_SGRBG12
    || pixelformat == V4L2_PIX_FMT_SGRBG12P
    || pixelformat == V4L2_PIX_FMT_SGRBG10
    || pixelformat == V4L2_PIX_FMT_SGRBG10P
    || pixelformat == V4L2_PIX_FMT_SGRBG8
    )
        return CV_BayerGR2RGB;
//    else if(
//       pixelformat == V4L2_PIX_FMT_SRGGB12
//    || pixelformat == V4L2_PIX_FMT_SRGGB12P
//    || pixelformat == V4L2_PIX_FMT_SRGGB10
//    || pixelformat == V4L2_PIX_FMT_SRGGB10P
//    || pixelformat == V4L2_PIX_FMT_SRGGB8
//    )
        return CV_BayerRG2RGB;
    
}



char *fourcc2string(int fourcc)
{
        pixelformat = fourcc;

        if(pixelformat == V4L2_PIX_FMT_SBGGR10
        || pixelformat == V4L2_PIX_FMT_SGBRG10
        || pixelformat == V4L2_PIX_FMT_SGRBG10
        || pixelformat == V4L2_PIX_FMT_SRGGB10
        || pixelformat == V4L2_PIX_FMT_SBGGR10P
        || pixelformat == V4L2_PIX_FMT_SGBRG10P
        || pixelformat == V4L2_PIX_FMT_SGRBG10P
        || pixelformat == V4L2_PIX_FMT_SRGGB10P){
            pixelbit = 10;
        }else if(
           pixelformat == V4L2_PIX_FMT_SBGGR12
        || pixelformat == V4L2_PIX_FMT_SGBRG12
        || pixelformat == V4L2_PIX_FMT_SGRBG12
        || pixelformat == V4L2_PIX_FMT_SRGGB12
        || pixelformat == V4L2_PIX_FMT_SBGGR12P
        || pixelformat == V4L2_PIX_FMT_SGBRG12P
        || pixelformat == V4L2_PIX_FMT_SGRBG12P
        || pixelformat == V4L2_PIX_FMT_SRGGB12P){
            pixelbit = 12;
        }

        int i;
        for(i=0; i<4; i++){
                str[i] = (char)(fourcc >> 8*i) & 0xff;
        }
        str[i] = '\0';
        return(str);
}



int xioctl (int fh, int request, void *arg)
{
    int r;

    do
    {
        r = ioctl(fh, request, arg);
    } while (-1 == r && EINTR == errno);

    return r;
}




void disp_qctl_head(void)
{
    printf("%-15s:",  "name");
    printf(" %-10s,", "id");
//    printf(" %-5s,",  "type");
    printf(" %-10s,",  "type");
    printf(" %-5s,",  "min");
    printf(" %-5s,",  "max");
    printf(" %-5s,",  "step");
    printf(" %-5s,",  "default");
//    printf(" %-10s,","flags");
    printf(" %-5s\n",  "value");
}



void disp_qctl(struct v4l2_queryctrl *qctl, struct v4l2_control *control){
        const char *str = qctl->type < V4L2_CTRL_COMPOUND_TYPES 
                  ? ctl_type1[qctl->type - V4L2_CTRL_TYPE_INTEGER]
                  : ctl_type2[qctl->type - V4L2_CTRL_COMPOUND_TYPES];
        printf("%-15s:",qctl->name);
        printf(" 0x%08X,",qctl->id);
//        printf(" %5d,",qctl->type);
        printf(" %-10s,",str);
        printf(" %5d,",qctl->minimum);
        printf(" %5d,",qctl->maximum);
        printf(" %5d,",qctl->step);
        printf(" %5d,",qctl->default_value);
//        printf(" 0x%08X",qctl->flags);
        printf(" %5d\n",control->value);
}



void set_parameter(int fd, int id, int value){
    struct v4l2_queryctrl qctl;
    struct v4l2_control control;

    memset(&qctl, 0, sizeof(qctl));
    qctl.id = id;
    if (-1 == xioctl(fd, VIDIOC_QUERYCTRL, &qctl)) {
        if (errno != EINVAL) {
            perror("VIDIOC_QUERYCTRL");
            exit(EXIT_FAILURE);
        } else {
            fprintf(stderr, "%s control is not surported\n", qctl.name);
        }
    } else if (qctl.flags & V4L2_CTRL_FLAG_DISABLED) {
        fprintf(stderr, "%s control is not surported\n", qctl.name);
    } else {
        memset(&control, 0, sizeof(control));
        control.id = id;
        xioctl(fd, VIDIOC_G_CTRL, &control);
        if(value < qctl.minimum)
                value = qctl.minimum;
        if(value > qctl.maximum)
                value = qctl.maximum;
        control.value = value;
        if (-1 == xioctl(fd, VIDIOC_S_CTRL, &control)) {
            perror("VIDIOC_S_CTRL");
            exit(EXIT_FAILURE);
        }
//        disp_qctl(&qctl, &control);
    }
}



void init_device(int fd){
        struct v4l2_capability cap;
        struct v4l2_cropcap cropcap;
        struct v4l2_crop crop;
        struct v4l2_format fmt;
        unsigned int min;

        if (-1 == xioctl(fd, VIDIOC_QUERYCAP, &cap)){
                printf("error VIDIOC_QUERYCAP\n");
                return;
        }else{
                printf("[device info]\n");
                printf("\tdriver  : %s\n",cap.driver);
                printf("\tcard    : %s\n",cap.card);
                printf("\tbus info: %s\n",cap.bus_info);
                printf("\tAvailable capabilities of the physical device: 0x%08X\n",cap.capabilities);
                printf("\t capabilities & V4L2_CAP_VIDEO_CAPTURE       : 0x%08X\n",cap.capabilities & V4L2_CAP_VIDEO_CAPTURE);
                printf("\t capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE: 0x%08X\n",cap.capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE);
                printf("\t capabilities & V4L2_CAP_READWRITE           : 0x%08X\n",cap.capabilities & V4L2_CAP_READWRITE);
                printf("\t capabilities & V4L2_CAP_STREAMING           : 0x%08X\n",cap.capabilities & V4L2_CAP_STREAMING);
                printf("\tDevice capabilities of the opened device     : 0x%08X\n",cap.device_caps);
        }
        printf("\n");

        CLEAR(cropcap);
        if(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE){
                cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        }else if(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE){
                cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
        }
        if(-1 == xioctl(fd, VIDIOC_CROPCAP, &cropcap)){
                printf("error VIDIOC_CROPCAP\n");
        }else{
                printf("cropcap\n");
                printf("\tleft  : %d\n", cropcap.defrect.left);
                printf("\ttop   : %d\n", cropcap.defrect.top);
                printf("\twidth : %d\n", cropcap.defrect.width);
                printf("\theight: %d\n", cropcap.defrect.height);
        }

        if(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE){
                crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        }else if(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE){
                crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
        }
        crop.c = cropcap.defrect; /* reset to default */
        if(-1 == xioctl(fd, VIDIOC_S_CROP, &crop)){
                printf("error VIDIOC_S_CROP\n");
        }else{
                printf("crop\n");
                printf("\tleft  : %d\n", crop.c.left);
                printf("\ttop   : %d\n", crop.c.top);
                printf("\twidth : %d\n", crop.c.width);
                printf("\theight: %d\n", crop.c.height);
        }
        printf("\n");

        CLEAR(fmt);

        if(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE){

                fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                fmt.fmt.pix.width       = WIDTH;
                fmt.fmt.pix.height      = HEIGHT;
//                fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_SGRBG8;
//                fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_SGRBG10P;
                fmt.fmt.pix.pixelformat = pixformat;
                fmt.fmt.pix.field       = V4L2_FIELD_ANY;

                if(-1 == xioctl(fd, VIDIOC_S_FMT, &fmt));

                printf("[set format]\n");
                printf("\ttype:   %d\n",fmt.type);
                printf("\twidth:  %d\n",fmt.fmt.pix.width);
                printf("\theight: %d\n",fmt.fmt.pix.height);
                printf("\n");

                if(-1 == xioctl(fd, VIDIOC_G_FMT, &fmt));
                //Buggy driver paranoia.
                min = fmt.fmt.pix.width * 2;
                if (fmt.fmt.pix.bytesperline < min)
                        fmt.fmt.pix.bytesperline = min;
                min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
                if (fmt.fmt.pix.sizeimage < min)
                        fmt.fmt.pix.sizeimage = min;

                printf("[get format]\n");
                printf("\ttype : 0x%08X\n", fmt.type);
                printf("\twidth : %d\n", fmt.fmt.pix.width);
                printf("\theight: %d\n", fmt.fmt.pix.height);
                printf("\tpixelformat: %s(0x%08X)\n", fourcc2string(fmt.fmt.pix.pixelformat),fmt.fmt.pix.pixelformat);
                printf("\tfield: %d\n", fmt.fmt.pix.field);
                printf("\tsizeimage: %d\n", fmt.fmt.pix.sizeimage);

                buffers = (struct buffer *)calloc(1, sizeof(*buffers));
                buffers[0].length = fmt.fmt.pix.sizeimage;
                buffers[0].start = malloc(fmt.fmt.pix.sizeimage);
        }else if(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE_MPLANE){

                fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
                fmt.fmt.pix_mp.width       = WIDTH;
                fmt.fmt.pix_mp.height      = HEIGHT;
//                fmt.fmt.pix_mp.pixelformat = V4L2_PIX_FMT_SGRBG8;
//                fmt.fmt.pix_mp.pixelformat = V4L2_PIX_FMT_SGRBG10P;
                fmt.fmt.pix_mp.pixelformat = pixformat;
                fmt.fmt.pix_mp.field       = V4L2_FIELD_ANY;

                if(-1 == xioctl(fd, VIDIOC_S_FMT, &fmt));

                printf("set format\n");
                printf("\ttype  : %d\n",fmt.type);
                printf("\twidth : %d\n",fmt.fmt.pix_mp.width);
                printf("\theight: %d\n",fmt.fmt.pix_mp.height);
                printf("\n");

                if(-1 == xioctl(fd, VIDIOC_G_FMT, &fmt));
                //Buggy driver paranoia.
                min = fmt.fmt.pix_mp.width * 2;
                if (fmt.fmt.pix_mp.plane_fmt[0].bytesperline < min)
                        fmt.fmt.pix_mp.plane_fmt[0].bytesperline = min;
                min = fmt.fmt.pix_mp.plane_fmt[0].bytesperline * fmt.fmt.pix_mp.height;
                if (fmt.fmt.pix_mp.plane_fmt[0].sizeimage < min)
                        fmt.fmt.pix_mp.plane_fmt[0].sizeimage = min;

                printf("[get format]\n");
                printf("\ttype  : %d\n", fmt.type);
                printf("\twidth : %d\n", fmt.fmt.pix_mp.width);
                printf("\theight: %d\n", fmt.fmt.pix_mp.height);
                printf("\tfield : %d\n", fmt.fmt.pix_mp.field);
                printf("\tsizeimage  : %d\n", fmt.fmt.pix_mp.plane_fmt[0].sizeimage);
                printf("\tpixelformat: %s(0x%08X)\n", fourcc2string(fmt.fmt.pix_mp.pixelformat));

                buffers = (struct buffer *)calloc(1, sizeof(*buffers));
                buffers[0].length = fmt.fmt.pix_mp.plane_fmt[0].sizeimage;
                buffers[0].start = malloc(fmt.fmt.pix_mp.plane_fmt[0].sizeimage);
        }
}



int read_frame(int fd){
        fd_set fds;
        struct timeval tv;
        int r;

        FD_ZERO(&fds);
        FD_SET(fd, &fds);

        // Timeout.
        tv.tv_sec = 10;
        tv.tv_usec = 0;

        r = select(fd + 1, &fds, NULL, NULL, &tv);
        if (-1 == r) {
                if (EINTR == errno)
                        return 0;
                fprintf(stderr, "error %d, %s\n", errno, strerror(errno));
                exit(EXIT_FAILURE);
        }

        if (0 == r) {
                fprintf(stderr, "select timeout\n");
                exit(EXIT_FAILURE);
        }

//        return read(fd, buffers[0].start, buffers[0].length);
        read_buffer_size = read(fd, buffers[0].start, buffers[0].length);
        return read_buffer_size;
}



void cp_buff_to_image(void){
        unsigned char *sp = (unsigned char *)(buffers[0].start);
        unsigned char *dp = (unsigned char *)(BayerImg->imageData);
 
        if(pixelformat == V4L2_PIX_FMT_SBGGR12
        || pixelformat == V4L2_PIX_FMT_SGBRG12
        || pixelformat == V4L2_PIX_FMT_SGRBG12
        || pixelformat == V4L2_PIX_FMT_SRGGB12){ //12bit
                for(int i=0, j=0; j<BayerImg->imageSize; i+=2,j++){
                        unsigned int tmp = ((unsigned int)sp[i] + (unsigned int)sp[i+1] * 0x100)/0x10;
                        dp[j] = (unsigned char)tmp;
                }
        }else if(pixelformat == V4L2_PIX_FMT_SBGGR12P
        || pixelformat == V4L2_PIX_FMT_SGBRG12P
        || pixelformat == V4L2_PIX_FMT_SGRBG12P
        || pixelformat == V4L2_PIX_FMT_SRGGB12P){ //Packed 12bit
                for(int i=0, j=0; j<BayerImg->imageSize; i++,j++){
                        if(i%3 == 2){
                                i++;
                        }
                        dp[j] = sp[i];
                }
        }else if(pixelformat == V4L2_PIX_FMT_SBGGR10
        || pixelformat == V4L2_PIX_FMT_SGBRG10
        || pixelformat == V4L2_PIX_FMT_SGRBG10
        || pixelformat == V4L2_PIX_FMT_SRGGB10){ //10bit
                for(int i=0, j=0; j<BayerImg->imageSize; i+=2,j++){
                        unsigned int tmp = ((unsigned int)sp[i] + (unsigned int)sp[i+1] * 0x100)/0x04;
                        dp[j] = (unsigned char)tmp;
//                for(int i=0, j=0; j<BayerImg->imageSize; i++,j++){
//                        dp[j] = sp[i];
                }
        }else if(pixelformat == V4L2_PIX_FMT_SBGGR10P
        || pixelformat == V4L2_PIX_FMT_SGBRG10P
        || pixelformat == V4L2_PIX_FMT_SGRBG10P
        || pixelformat == V4L2_PIX_FMT_SRGGB10P){ //Packed 10bit
                for(int i=0, j=0; j<BayerImg->imageSize; i++,j++){
                        if(i%5 == 4){
                                i++;
                        }
                        dp[j] = sp[i];
                }
        }else{
//        if(pixelformat == V4L2_PIX_FMT_SBGGR8
//        || pixelformat == V4L2_PIX_FMT_SGBRG8
//        || pixelformat == V4L2_PIX_FMT_SGRBG8
//        || pixelformat == V4L2_PIX_FMT_SRGGB8){ //8bit
                for(int i=0, j=0; j<BayerImg->imageSize; i++,j++){
                        dp[j] = sp[i];
                }
        }

}



void find_ctl(int fd){
    struct v4l2_queryctrl qctl;
    struct v4l2_control control;

    memset(&qctl, 0, sizeof(qctl));
    controls=0;
    for(int i=0; 0 != ctl_ids[i]; i++){
        qctl.id = ctl_ids[i];
        if (-1 == xioctl(fd, VIDIOC_QUERYCTRL, &qctl)) {
            if (errno != EINVAL) {
                perror("VIDIOC_QUERYCTRL");
                exit(EXIT_FAILURE);
            }
        } else if (qctl.flags & V4L2_CTRL_FLAG_DISABLED) {
//            printf("%s control is not surported\n", qctl.name);
        } else {
            control.id = ctl_ids[i];
            xioctl(fd, VIDIOC_G_CTRL, &control);
            controls++;
        }
    }
    printf("\n");

    effective_ctl = (int *)malloc(sizeof(int) * controls);
    for(int i=0, j=0; 0 != ctl_ids[i]; i++){
        qctl.id = ctl_ids[i];
        if (-1 != xioctl(fd, VIDIOC_QUERYCTRL, &qctl)) 
            if (!qctl.flags & V4L2_CTRL_FLAG_DISABLED)
               effective_ctl[j++] = ctl_ids[i]; //register effective ctls
    }
    ctl_mode_index = 0;
    ctl_mode = effective_ctl[ctl_mode_index];
}

int get_parameter(int fd, int id){
    struct v4l2_queryctrl qctl;
    struct v4l2_control control;

    memset(&qctl, 0, sizeof(qctl));
    memset(&control, 0, sizeof(control));
    qctl.id = id;
    if (-1 == xioctl(fd, VIDIOC_QUERYCTRL, &qctl)) {
        if (errno != EINVAL) {
            perror("VIDIOC_QUERYCTRL");
            exit(EXIT_FAILURE);
        }
    } else if (qctl.flags & V4L2_CTRL_FLAG_DISABLED) {
//        fprintf(stderr, "%s control is not surported\n", qctl.name);
    } else {
        control.id = id;
        xioctl(fd, VIDIOC_G_CTRL, &control);
        disp_qctl(&qctl, &control);
    }
    return control.value;
}


void list_parameter(int fd){
    printf("[effective control list]\n\tkey: ");
    disp_qctl_head();

    for(int i=0; i < controls; i++){
        printf("\t%2d : ",i);
        get_parameter(fd, effective_ctl[i] );
    }
    printf("\n");
}


void save_image_raw(void){
    char fname[64]="";

    time_t now = time(NULL);
    struct tm *pnow = localtime(&now);
    sprintf(fname, "%04d-%02d-%02d_%02d_%02d_%02d_%dx%d_%s.raw",
            pnow->tm_year+1900, pnow->tm_mon+1, pnow->tm_mday,
            pnow->tm_hour, pnow->tm_min, pnow->tm_sec, WIDTH, HEIGHT, str);

    FILE *file;
    file = fopen(fname,"wb");
    if(file){
        fwrite(buffers[0].start, read_buffer_size, 1, file);
        fclose(file);
        printf("\nSave image file: %s\n", fname);
    }
}


void save_image_bmp(void){
    IplImage *Img = RGBImg;
    char fname[64]="";

    time_t now = time(NULL);
    struct tm *pnow = localtime(&now);
    sprintf(fname, "%04d-%02d-%02d_%02d_%02d_%02d_%dx%d_%s.bmp",
            pnow->tm_year+1900, pnow->tm_mon+1, pnow->tm_mday,
            pnow->tm_hour, pnow->tm_min, pnow->tm_sec, WIDTH, HEIGHT, str);

    printf("\nSave image file: %s\n", fname);
    cvSaveImage(fname, RGBImg, 0);
}


void save_image_jpg(void){
    IplImage *Img = RGBImg;
    char fname[64]="";

    time_t now = time(NULL);
    struct tm *pnow = localtime(&now);
    sprintf(fname, "%04d-%02d-%02d_%02d_%02d_%02d_%dx%d_%s.jpg",
            pnow->tm_year+1900, pnow->tm_mon+1, pnow->tm_mday,
            pnow->tm_hour, pnow->tm_min, pnow->tm_sec, WIDTH, HEIGHT, str);

    printf("\nSave image file: %s\n", fname);
    cvSaveImage(fname, RGBImg, 0);
}


// Key Code
#define KSHIFT 0x01
#define KCTRL  0x04
#define KALT   0x08
#define KLEFT  0x51
#define KUP    0x52
#define KRIGHT 0x53
#define KDOWN  0x54
#define KESC   0x1B

int ProcKey(int fd){
    int KeyCode = cvWaitKey(1);
    int SpecialCode = KeyCode / 0x10000;
    KeyCode &= 0x00ff;

    if(KeyCode == 'q' || KeyCode == KESC){
        printf("Quit\n");
        return 1;
    }else if(KeyCode == 'r'){   // raw
        save_image_raw();
    }else if(KeyCode == 'b'){   // bmp
        save_image_bmp();
    }else if(KeyCode == 'j'){   // jpeg
        save_image_jpg();
    }else if(KeyCode >= '0' && KeyCode <= '9'){   // mode select
        ctl_mode_index = KeyCode - '0';
        if(ctl_mode_index < controls){
            ctl_mode = effective_ctl[ctl_mode_index];
            printf("\npress %c key\n", KeyCode);
            get_parameter(fd, ctl_mode);
        }
    }else if(KeyCode == KRIGHT || KeyCode == KUP
          || KeyCode == KLEFT  || KeyCode == KDOWN){
        int incremental = 1;

        switch(SpecialCode){
         case KALT  : incremental *= 5;
         case KCTRL : incremental *= 5;
         case KSHIFT: incremental *= 5;
         default:    break;
        }
        if(KeyCode == KLEFT || KeyCode == KDOWN) incremental *= -1;

        if(KeyCode == KRIGHT || KeyCode == KLEFT){  // mode select
            ctl_mode_index += incremental;
            if(ctl_mode_index < 0) ctl_mode_index = controls - 1;
            else if(ctl_mode_index > controls) ctl_mode_index = 0;
            ctl_mode = effective_ctl[ctl_mode_index];
            printf("\npress 0x%X key\n", KeyCode);
            get_parameter(fd, ctl_mode);
        }else if(KeyCode == KUP || KeyCode == KDOWN){   // value change
            set_parameter(fd, ctl_mode, get_parameter(fd, ctl_mode) + incremental);
        }
    }
    return 0;
}

void pixformat_list(void){
    int pixfotmat[]={
        V4L2_PIX_FMT_BGR24,
        V4L2_PIX_FMT_BGR24,
        V4L2_PIX_FMT_SBGGR8,
        V4L2_PIX_FMT_SGBRG8,
        V4L2_PIX_FMT_SGRBG8,
        V4L2_PIX_FMT_SRGGB8,
        V4L2_PIX_FMT_SBGGR10,
        V4L2_PIX_FMT_SGBRG10,
        V4L2_PIX_FMT_SGRBG10,
        V4L2_PIX_FMT_SRGGB10,
        V4L2_PIX_FMT_SBGGR10P,
        V4L2_PIX_FMT_SGBRG10P,
        V4L2_PIX_FMT_SGRBG10P,
        V4L2_PIX_FMT_SRGGB10P,
        V4L2_PIX_FMT_SBGGR12,
        V4L2_PIX_FMT_SGBRG12,
        V4L2_PIX_FMT_SGRBG12,
        V4L2_PIX_FMT_SRGGB12,
        V4L2_PIX_FMT_SBGGR12P,
        V4L2_PIX_FMT_SGBRG12P,
        V4L2_PIX_FMT_SGRBG12P,
        V4L2_PIX_FMT_SRGGB12P
    };
 
    printf("  pixformat:");
    for(int i=0; i<sizeof(pixfotmat)/sizeof(int); i++){
        printf("%s ",fourcc2string(pixfotmat[i]));
    }
    printf("\n");
}


unsigned long stofcc(char *cp){
    unsigned long fcc = 0;
    for(int i=0; i<4; i++,cp++){
//        printf("%c",*cp);
        fcc += (unsigned long)(*cp << 8*i);
    }
//    printf("\n%x\n",fcc);
    return fcc;
}

int main(int argc, char *argv[]){

        int fd;
        int status;
        struct stat st;
        if(argc != 7){
                printf("usage:%s [pixformat] [width] [height] [analog gain] [vblank] [exposure]\n",argv[0]);
                pixformat_list();
                exit(-1);
        }
        pixformat    = stofcc(argv[1]);
        WIDTH        = atoi(argv[2]);
        HEIGHT       = atoi(argv[3]);
        int aGain    = atoi(argv[4]);
        int vblank    = atoi(argv[5]);
        int exposure = atoi(argv[6]);
//        printf("request format:%s\n",fourcc2string(pixformat));
//        printf("request size:%dx%d\n\n",WIDTH, HEIGHT);

        cvNamedWindow("Viewer", CV_WINDOW_AUTOSIZE);
        BayerImg = cvCreateImage(cvSize(WIDTH,HEIGHT),IPL_DEPTH_8U,1);
//        BayerImg = cvCreateImage(cvSize(WIDTH,HEIGHT),IPL_DEPTH_16U,1);
        RGBImg   = cvCreateImage(cvSize(WIDTH,HEIGHT),IPL_DEPTH_8U,3);

        status = stat(dev_name, &st);
        printf("device name:%s\n",dev_name);
        printf("\tstat:%d\n",status);
        printf("\tst_dev:%d, st_rdev:%d\n",st.st_dev, st.st_rdev);

        fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
        printf("\tfd:%d\n\n",fd);

        init_device(fd);
        find_ctl(fd);
        set_parameter(fd, V4L2_CID_ANALOGUE_GAIN, aGain);        //Analog Gain set
        set_parameter(fd, V4L2_CID_VBLANK, vblank);         //Digital Gain set
        set_parameter(fd, V4L2_CID_EXPOSURE, exposure);          //Exposure set
        list_parameter(fd);
        int cv_code = get_code();

        struct timeval tv_start, tv_end;
        struct timezone tz_local;
        gettimeofday(&tv_start, &tz_local);
        printf("[start capture]\n");
        while(1){
                int b = read_frame(fd);
                if(b > 0){
                        double fps;
                        if(frame_number%10 == 0){
                                gettimeofday(&tv_end, &tz_local);
                                double start_time = tv_start.tv_sec+(double)tv_start.tv_usec/1000000;
                                double end_time = tv_end.tv_sec+(double)tv_end.tv_usec/1000000;
                                fps = 10/(end_time-start_time);
                                gettimeofday(&tv_start, &tz_local);
                        }
                        printf("%d: read size: %d, framerate: %3.2ffps \r", frame_number++, b, fps);
                        cp_buff_to_image();
                        if(pixelformat == V4L2_PIX_FMT_BGR24 || pixelformat == V4L2_PIX_FMT_BGR24){
                            cvShowImage("Viewer",BayerImg);
                        }else{
                            cvCvtColor(BayerImg, RGBImg, cv_code);
                            cvShowImage("Viewer",RGBImg);
                        }
                        if(ProcKey(fd))
                                break;
                }
        }
        close(fd);
        printf("\n");
}
