/*====================================================================*
 * util.c
 *
 * Copyright(c) 2020-2021 Soho Enterprise Ltd. (https://soho-enterprise.com/)
 *====================================================================*/
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/statvfs.h>

static const char __PROGRAM_VERSION__[] = "@(#)util.c(SensorLaunchRAW) Ver. 1.0005 (2021/05/11 11:41:34)";

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
GetDiskSize(const char *diskname, long *size, long *free)
{
    struct statvfs vfs;

    if (statvfs(diskname, &vfs) < 0) {
        return  -1;
    }
    if (vfs.f_fsid != 0) {  // Not found "/mnt/ram"
        return  -1;
    }
    *size = vfs.f_bsize * vfs.f_blocks;
    *free = vfs.f_bsize * vfs.f_bfree;
    return  0;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
double
GetTime()
{
    struct  timeval t;

    gettimeofday(&t, NULL);

    return  (double)t.tv_sec + (double)t.tv_usec / 1000000.0;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
char *
CreateDateTime()
{
    static  char    date_time[64];
    time_t  stime;
    struct  tm *now;
    struct  timeval tv;
    int     yy, mm, dd, h, m, s;
     
    stime = time(NULL);
    now = localtime(&stime);
    gettimeofday(&tv, NULL);
 
    yy = (now->tm_year + 1900) % 2000;
    mm = now->tm_mon + 1;
    dd = now->tm_mday;
    h  = now->tm_hour;
    m  = now->tm_min;
    s  = now->tm_sec;

    sprintf(date_time, "%02d%02d%02d_%02d%02d%02d_%06d", yy, mm, dd, h, m, s, tv.tv_usec);

    return date_time;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
char *
MakeFileName(int width, int height)
{
    static  char    fname[BUFSIZ];

    sprintf(fname, "%dx%d_%s", width, height, CreateDateTime());

    return  fname;
}

/*--------------------------------------------------------------------*
 *
 *--------------------------------------------------------------------*/
int
WriteRAW(char *fname, unsigned char *raw, int size)
{
    FILE    *fp;

    if ((fp = fopen(fname, "wb")) == NULL) {
        printf("can't open %s\n", fname);
        return  -1;
    }
    if (fwrite(raw, size, 1, fp) != 1) {
        fclose(fp);
        return  -1;
    }
    fclose(fp);

    return  0;
}
