#include <stdio.h>
#include <opencv2/opencv.hpp>
#include <opencv2/highgui/highgui.hpp>
#include "zoom.h"

cv::Point from, to;             // 画像切り取り領域開始終了点
cv::Point mfrom, mto;           // マウスで選択領域開始終了点(選択モード用)
cv::Point rfrom, rto;           // 画像切り取り領域開始終了点(拡大モード用)
cv::Size dSize, zSize;          // 表示ウィンドウ上のサイズ、カメラからの入力画像のサイズ
int ZoomMode    = ZOOM_INIT;
int ZoomModeOld = ZOOM_INIT;
float reductionRate = 1.0;


/*
 * マウスイベントで呼び出される関数
 *
 * 入力：マウスイベント、ウィンドウ内のX座標、ウィンドウ内のY座標
 */
void zoom_mouse_callback(int event, int x, int y, int flags, void *param){
    if(event == cv::EVENT_RBUTTONDOWN){
        if(ZoomMode == ZOOM_ON)
            cv::destroyWindow ("zoom"); // ウィンドウ消去
        ZoomMode = ZoomModeOld = ZOOM_INIT;

    }else if(event == cv::EVENT_LBUTTONDOWN){
        mfrom = cv::Point(x, y);
        mto   = cv::Point(x, y);
#ifdef _DEBUG
        printf("\nDisplay area Width:%d Height:%d\n", dSize.width, dSize.height);
        printf("From x:%03d y:%03d\n", mfrom.x, mfrom.y);
#endif
        ZoomModeOld = ZoomMode;
        ZoomMode = ZOOM_SELECT;

    }else if((event == cv::EVENT_MOUSEMOVE) && ZoomMode == ZOOM_SELECT){
        mto = cv::Point(x, y);
#ifdef _DEBUG
        printf("To   x:%03d y:%03d\n", mto.x, mto.y);
#endif

        // x座標補正
        if(mto.x < 0){                    // カーソルがウィンドウの左に出た場合は止めておく
            mto.x = 0;
        }else if(mto.x > dSize.width){    // カーソルがウィンドウの右に出た場合は止めておく
            mto.x = dSize.width;
        }
        if(mto.x < mfrom.x){              // 右から選択開始、左で選択終了の場合は入れ替えて使用
            to.x   = mfrom.x;
            from.x = mto.x;
        }else{                            // 左から選択開始、右で選択終了の場合はそのまま使用
            from.x = mfrom.x;
            to.x   = mto.x;
        }

        // y座標補正
        if(mto.y < 0){                    // カーソルがウィンドウの上に出た場合は止めておく
            mto.y = 0;
        }else if(mto.y > dSize.height){   // カーソルがウィンドウの下に出た場合は止めておく
            mto.y = dSize.height;
        }
        if(mto.y < mfrom.y){              // 下から選択開始、上で選択終了の場合は入れ替えて使用
            to.y   = mfrom.y;
            from.y = mto.y;
        }else{                            // 上から選択開始、下で選択終了の場合はそのまま使用
            from.y = mfrom.y;
            to.y   = mto.y;
        }

        from *= int(1.0/reductionRate);   // 1行にできない
        to   *= int(1.0/reductionRate);   //
        zSize = to - from;                //

    }else if(event == cv::EVENT_LBUTTONUP && ZoomMode == ZOOM_SELECT){
        if(x - mfrom.x == 0 || y - mfrom.y == 0){
            ZoomMode = ZoomModeOld;       // 領域面積が無い場合は、初期モードか拡大モードに戻す
            return;
        }
#ifdef _DEBUG
        printf("Select area Width:%d Height:%d\n", zSize.width, zSize.height);
#endif
        rfrom = mfrom;                    // 拡大領域を確定
        rto = mto;                        //

        if(ZoomModeOld == ZOOM_ON)
            cv::destroyWindow ("zoom");   // ウィンドウを一旦消去すると表示のタイミングで最前面に来る

        ZoomModeOld = ZoomMode;
        ZoomMode = ZOOM_ON;
    }
}


/*
 * 専用マウスイベント関数(zoom_mouse_callback)登録関数
 *
 * 入力：全体表示ウィンドウ、縮小率
 */
void zoom_setMouseCallback(char *title, float rate){
    reductionRate = rate;                                                                   // マウス関数登録に渡す引数として使用
    cv::setMouseCallback(title, zoom_mouse_callback, (void *)&reductionRate);               // 拡大用マウスイベンと処理を登録
}


/*
 * 拡大画像を表示し、元画像側に矩形を描画
 *
 * 入力：元画像、表示画像
 */
void zoom_imshow(cv::Mat srcMat, cv::Mat dspMat, float Rate){
    cv::Mat dsp8BitMat;
    if(ZoomMode == ZOOM_ON){
        cv::Rect roi(from, zSize);
        cv::resize(srcMat(roi), dsp8BitMat, cv::Size(), Rate, Rate);
        cv::imshow("zoom", dsp8BitMat);                                        // ウィンドウに画像を表示する
    }
    if(ZoomMode == ZOOM_ON){
        cv::rectangle(dspMat, rfrom, rto, CV_RGB(0,0,0), 4, 10, 0);             // 線と同色の場合の対応
        cv::rectangle(dspMat, rfrom, rto, CV_RGB(0,255,0), 2, 10, 0);           // 緑の矩形
    }else if(ZoomMode == ZOOM_SELECT){
        cv::rectangle(dspMat, mfrom, mto, CV_RGB(0,0,0), 4, 10, 0);             // 線と同色の場合の対応
        cv::rectangle(dspMat, mfrom, mto, CV_RGB(255,0,0), 2, 10, 0);           // 赤の矩形
    }
    dSize = dspMat.size();
}

